from setuptools import setup
import os

def read(*rnames):
        return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

setup(
    name='lucene-querybuilder',
    version='0.1.4',
    author='Edd Purcell',
    author_email='epurcell3@gatech.edu',
    maintainer='Matt Luongo',
    maintainer_email='mhluongo@gmail.com',
    description='A DSL to build Lucene text queries in Python.',
    url = "http://packages.python.org/lucene-querybuilder",
    packages=['lucenequerybuilder',],
    include_package_data=True,
    zip_safe=False,
    long_description=read('README.rst'),
    platforms=['posix'],
    tests_require=[
        'nose>=1.0',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Topic :: Text Processing :: Indexing',
        'Programming Language :: Python'
    ]
)



"""
Simple tests for Q. In the future, comparing output to an actual Lucene index
would be a good idea.
"""

from lucenequerybuilder import Q

def test_general():
    a = 'a'
    b = 'b'
    c = 'c'
    d = 'd d'
    e = 'e'
    f = 'f'
    g = 1
    h = 5
    q1 = Q(a) | Q(b)
    assert str(q1) == '('+a+' OR '+b+')', str(q1)
    q2 = Q(c) & q1
    assert str(q2) == '('+c+' AND ('+a+' OR '+b+'))'
    q3 = +Q(a) +Q(b) | Q(f, Q(c) & ~Q(Q(d) | Q(e)))
    assert str(q3) == '+'+a+' +'+b+' '+f+':(('+c+' AND NOT '+'("'+d+'" OR '+e+')))'
    q4 = Q(a,b) | Q(c, Q(d) & Q(e)) | Q(f, +Q(inrange=(g,h)) | Q(h))
    assert str(q4) == a+':('+b+') '+c+':(("'+d+'" AND '+e+')) '+f+':(+['+str(g)+' TO '+str(h)+'] '+str(h)+')'
    try:
        q5 = Q(a, Q(b, Q(c, d) & Q(e)) | Q(f))
    except:
        pass
    else:
        raise AssertionError("Shouldn't allow nested fields.")

def test_simple_term():
    query_string = str(Q('a'))
    assert query_string == 'a', query_string

def test_simple_phrase():
    query_string = str(Q('abc 123'))
    assert query_string == '"abc 123"', query_string

def test_hashing():
    q1 = Q('a') & Q('b') | Q('c')
    q2 = Q('a') & Q('b') | Q('c')
    q3 = q1 | Q('d')

    assert q1 == q2, "Queries aren't being properly evaluated for equality."
    assert q2 != q3, "Queries aren't being properly evaluated for inequality."

    d = {}
    try:
        d[q1] = 1
        d[q2] = 2
    except:
        raise AssertionError('There was an error using queries as dict keys.')
    assert d[q2] == 2, 'Got the wrong value back from the query dict!'

    q4 = Q('field', inrange=(1,2))
    q5 = Q('field', inrange=(1,3))
    assert q4 != q5
    assert hash(q4) != hash(q5)

def test_field_restrictions():
    q1 = Q('field', 'test query')
    assert q1.fielded

    q2 = Q('a') & q1
    assert q2.fielded

    try:
        Q('another_field', q2)
    except:
        pass
    else:
        raise AssertionError('Query allowed nested fields, which are invalid.')



class Q(object):
    """ Q is a query builder for the lucene language."""

    specialchars = r'+-!(){}[]^"~*?\:'
    doublechars = '&&||'

    def __init__(self, *args, **kwargs):
        """
        """
        self.must = []
        self.must_not = []
        self.should = []
        self._and = None
        self._or = None
        self._not = None
        self.inrange = None
        self.exrange = None
        self._child_has_field = False
        self.field = None
        if len(args) == 1 and not kwargs:
            if isinstance(args[0], Q):
                if args[0].fielded:
                    self._child_has_field = True
                self.should.append(args[0])
            else:
                if Q._check_whitespace(args[0]):
                    self.should.append('"'+self._escape(args[0])+'"')
                else:
                    self.should.append(self._escape(args[0]))
        elif len(args) <= 1 and kwargs:
            if kwargs.get('inrange'):
                self.inrange = tuple(kwargs['inrange'])
            elif kwargs.get('exrange'):
                self.exrange = tuple(kwargs['exrange'])
            if len(args) == 1:
                if Q._check_whitespace(args[0]):
                    raise ValueError('No whitepsace allowed in field names.')
                self.field = args[0]
        elif len(args) == 2:
            if Q._check_whitespace(args[0]):
                raise ValueError('No whitespace allowed in field names.')
            self.field = args[0]
            if isinstance(args[1], Q):
                if args[1].fielded:
                    raise ValueError('No nested fields allowed.')
                self.should.append(args[1])
            else:
                if Q._check_whitespace(args[1]):
                    self.should.append('"'+self._escape(args[1])+'"')
                else:
                    self.should.append(self._escape(args[1]))

    @property
    def fielded(self):
        return self.field is not None or\
                any(Q._has_field(l) for l in [self.must, self.must_not,
                                              self.should, self._and, self._or,
                                              self._not])
    @staticmethod
    def _has_field(val):
        if hasattr(val, '__iter__'):
            return any(Q._has_field(t) for t in val)
        else:
            return hasattr(val, 'field') and val.field is not None


    @classmethod
    def _check_whitespace(cls, s):
        import string
        if isinstance(s, basestring):
            for c in string.whitespace:
                if c in s:
                    return True
        return False

    @classmethod
    def _escape(cls, s):
        if isinstance(s, basestring):
            rv = ''
            for c in s:
                if c in cls.specialchars:
                    rv += '\\' + c
                else:
                    rv += c
            return rv
        return s

    def _make_and(q1, q2):
        q = Q()
        q._and = (q1, q2)
        return q

    def _make_not(q1):
        q = Q()
        q._not = q1
        return q

    def _make_or(q1, q2):
        q = Q()
        q._or = (q1, q2)
        return q

    def _make_must(q1):
        q = Q()
        q.must.append(q1)
        return q

    def _make_must_not(q1):
        q = Q()
        q.must_not.append(q1)
        return q

    def __and__(self, other):
        return Q._make_and(self, other)

    def __or__(self, other):
        return Q._make_or(self, other)

    def __invert__(self):
        return Q._make_not(self)

    def __pos__(self):
        return Q._make_must(self)

    def __neg__(self):
        return Q._make_must_not(self)

    def __add__(self, other):
        return self | Q._make_must(other)

    def __sub__(self, other):
        return self | Q._make_must_not(self)

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __hash__(self):
        return hash((tuple(self.should), tuple(self.must),tuple(self.must_not),
                     self.exrange, self.inrange, self.field))

    def __str__(self):
        rv = ''
        for o in self.must:
            rv += '+' + str(o)
        for o in self.must_not:
            rv += str(o)
        for o in self.should:
            rv += str(o)
        if self._and is not None:
            rv += '(' + str(self._and[0]) + ' AND ' + str(self._and[1]) + ')'
        if self._not is not None:
            rv += 'NOT ' + str(self._not)
        if self._or is not None:
            if self._or[0].field is not None or self._or[1].field is not None\
               or self._or[0].must or self._or[1].must or self._or[0].must_not\
               or self._or[1].must_not:
                rv += str(self._or[0]) + ' ' + str(self._or[1])
            else:
                rv += '(' + str(self._or[0]) + ' OR ' + str(self._or[1]) + ')'
        if self.inrange is not None:
            rv += '[' + str(self.inrange[0]) + ' TO ' + str(self.inrange[1]) + ']'
        if self.exrange is not None:
            rv += '{' + str(self.exrange[0]) + ' TO ' + str(self.exrange[1]) + '}'
        if self.field is not None:
            rv = '{0}:({1})'.format(self.field, rv)
        return rv

    def _check_nested_fields(self):
        stack = []
        stack.extend(self.must)
        stack.extend(self.must_not)
        stack.extend(self.should)
        if hasattr(self, '_and'):
            stack.extend(self._and)
        if hasattr(self, '_not'):
            stack.append(self._not)
        if hasattr(self, '_or'):
            stack.extend(self._or)
        while stack:
            o = stack.pop()
            if not isinstance(o, Q):
                continue
            if hasattr(o, 'field') or o._child_has_field:
                self._child_has_field = True
                if self._has_field:
                    return True
            else:
                stack.extend(o.must)
                stack.extend(o.must_not)
                stack.extend(o.should)
                if hasattr(o, '_and'):
                    stack.extend(o._and)
                if hasattr(o, '_not'):
                    stack.append(o._not)
                if hasattr(o, '_or'):
                    stack.extend(o._or)
        return False
           
    #probably dfs here, build a flag tree as you go so dfs won't be run 100 times

